@@ -7,11 +7,15 @@
 
 from motorcortex.reply import Reply
 
+import os
+import ctypes
+
 from concurrent.futures import ThreadPoolExecutor
 from nanomsg import wrapper as nn
 from nanomsg import REQ, SOL_SOCKET, RCVTIMEO, RCVMAXSIZE, AF_SP, create_writable_buffer
 from builtins import bytes
 
+
 class Request(object):
     """Request/Reply communication is used to send commands to a motorcortex server.
 
@@ -62,18 +66,6 @@ class Request(object):
         self.__socket = None
         print("Reply connection closed")
 
-    def __send(self, encoded_msg, do_not_decode_reply):
-        socket = self.__socket
-        nn.nn_send(socket, encoded_msg, 0)
-        length, buffer = nn.nn_recv(socket, 0)
-        if length > 0:
-            if do_not_decode_reply:
-                return buffer
-            else:
-                return self.__protobuf_types.decode(bytes(buffer))
-
-        return None
-
     def send(self, encoded_msg, do_not_decode_reply=False):
         """Sends an encoded request to the server
 
@@ -122,12 +114,31 @@ class Request(object):
 
         return self.send(self.__protobuf_types.encode(login_msg))
 
+    def getParameterTreeHash(self):
+        """Requests a parameter tree hash from the server.
+
+            Returns:
+                Reply(ParameterTreeMsg): A Promise, which resolves when parameter tree hash is received or fails
+                otherwise. ParameterTreeHashMsg message has a status field to check the status of the operation.
+
+            Examples:
+                >>> param_tree_hash_reply = req.getParameterTreeHash()
+                >>> value = param_tree_hash_reply.get()
+
+        """
+
+        # getting and instantiating data type from the loaded dict
+        param_tree_hash_msg = self.__protobuf_types.createType('motorcortex.GetParameterTreeHashMsg')
+
+        # encoding and sending data
+        return self.send(self.__protobuf_types.encode(param_tree_hash_msg))
+
     def getParameterTree(self):
         """Requests a parameter tree from the server.
 
             Returns:
                 Reply(ParameterTreeMsg): A Promise, which resolves when parameter tree is received or fails
-                otherwise. ParameterTreeMsg message has a status files to check the status of the operation.
+                otherwise. ParameterTreeMsg message has a status field to check the status of the operation.
 
             Examples:
                 >>> param_tree_reply = req.getParameterTree()
@@ -136,11 +147,22 @@ class Request(object):
 
         """
 
+        hash = self.getParameterTreeHash().get()
+        tree = self.__loadParameterTreeFile(str(hash.hash))
+        if self.__loadParameterTreeFile(str(hash.hash)):
+            if hash.hash == self.__hash(tree):
+                print('Found parameter tree in the cache')
+                return Reply(self.__pool.submit(lambda tree: tree, tree))
+            else:
+                print('Found parameter tree but the cache does not match')
+        else:
+            print('Failed to find parameter tree in the cache')
+
         # getting and instantiating data type from the loaded dict
         param_tree_msg = self.__protobuf_types.createType('motorcortex.GetParameterTreeMsg')
-
+        handle = self.send(self.__protobuf_types.encode(param_tree_msg))
         # encoding and sending data
-        return self.send(self.__protobuf_types.encode(param_tree_msg))
+        return Reply(self.__pool.submit(lambda handle: self.__saveParameterTreeFile(handle.get()), handle))
 
     def save(self, path, file_name):
         """Request a server to save a parameter tree to file.
@@ -320,3 +342,55 @@ class Request(object):
         get_param_msg.path = path
 
         return get_param_msg
+
+    def __hash(self, tree):
+        hash = int()
+        for param in tree.params:
+            clip_u32 = lambda val: ctypes.c_uint32(val).value
+            hash += clip_u32(param.data_size + param.data_type + param.id + param.number_of_elements + \
+                             param.param_type + param.unit + param.group_id + param.permissions)
+            for ch in param.path:
+                hash = clip_u32(((hash << 5) - hash) + ord(ch))
+                hash |= 0
+
+            if hasattr(param, 'module_type'):
+                for ch in param.module_type:
+                    hash = (clip_u32(hash << 3) - hash) + ord(ch)
+
+        return hash
+
+    def __loadParameterTreeFile(self, path):
+        if os.path.exists(path):
+            text_file = open(path, "rb")
+            data = text_file.read()
+            text_file.close()
+            param_tree_hash_msg = self.__protobuf_types.createType('motorcortex.ParameterTreeMsg')
+            param_tree_hash_msg.ParseFromString(data)
+            return param_tree_hash_msg
+        return None
+
+    def __saveParameterTreeFile(self, parameter_tree):
+        print('Saved parameter tree to the cache')
+        cache = self.__hash(parameter_tree)
+        path = str(cache)
+        data = parameter_tree.SerializeToString()
+        text_file = open(path, "wb")
+        text_file.write(data)
+        text_file.close()
+        return parameter_tree
+
+    def __cachedTree(self, tree):
+        return tree
+
+    def __send(self, encoded_msg, do_not_decode_reply):
+        socket = self.__socket
+        nn.nn_send(socket, encoded_msg, 0)
+        length, buffer = nn.nn_recv(socket, 0)
+        if length > 0:
+            if do_not_decode_reply:
+                return buffer
+            else:
+                return self.__protobuf_types.decode(bytes(buffer))
+
+        return None
+
